"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.addFixedValues = exports.dynamicItemsOfLayout = exports.fixedItemsOfLayout = void 0;
const layout_1 = require("./layout");
const fixedItemsOfLayout = (layout) => layout.reduce((acc, item) => {
    if (item["custom"] !== undefined && ((0, layout_1.isPrimitiveType)(item["custom"]) || (0, layout_1.isPrimitiveType)(item["custom"].from)))
        return [...acc, item];
    if (item.binary === "array" || item.binary === "object") {
        const fixedItems = (0, exports.fixedItemsOfLayout)(item.layout);
        if (fixedItems.length > 0)
            return [...acc, { ...item, layout: fixedItems }];
    }
    return acc;
}, []);
exports.fixedItemsOfLayout = fixedItemsOfLayout;
const dynamicItemsOfLayout = (layout) => layout.reduce((acc, item) => {
    if (item["custom"] === undefined || !((0, layout_1.isPrimitiveType)(item["custom"]) || (0, layout_1.isPrimitiveType)(item["custom"].from)))
        return [...acc, item];
    if (item.binary === "array" || item.binary === "object") {
        const dynamicItems = (0, exports.dynamicItemsOfLayout)(item.layout);
        if (dynamicItems.length > 0)
            return [...acc, { ...item, layout: dynamicItems }];
    }
    return acc;
}, []);
exports.dynamicItemsOfLayout = dynamicItemsOfLayout;
const addFixedValues = (layout, dynamicValues) => {
    const ret = {};
    for (const item of layout) {
        if (item.binary === "object") {
            const subDynamicValues = ((item.name in dynamicValues)
                ? dynamicValues[item.name]
                : {});
            ret[item.name] = (0, exports.addFixedValues)(item.layout, subDynamicValues);
        }
        else if (item.binary === "array") {
            const subDynamicValues = ((item.name in dynamicValues)
                ? dynamicValues[item.name]
                : []);
            ret[item.name] = subDynamicValues.map(element => (0, exports.addFixedValues)(item.layout, element));
        }
        else if (item.custom !== undefined &&
            ((0, layout_1.isPrimitiveType)(item.custom) || (0, layout_1.isPrimitiveType)(item.custom.from))) {
            if (!item?.omit)
                ret[item.name] = (0, layout_1.isPrimitiveType)(item.custom) ? item.custom : item.custom.to;
        }
        else
            ret[item.name] = dynamicValues[item.name];
    }
    return ret;
};
exports.addFixedValues = addFixedValues;
//# sourceMappingURL=fixedDynamic.js.map