export type PrimitiveType = number | bigint | Uint8Array;
export declare const isPrimitiveType: (x: any) => x is PrimitiveType;
export type BinaryLiterals = "uint" | "bytes" | "array" | "object";
export type NumberSize = 1 | 2 | 3 | 4 | 5 | 6;
export declare const numberMaxSize = 6;
export type UintSizeToPrimitive<Size extends number> = Size extends NumberSize ? number : bigint;
export type FixedConversion<FromType extends PrimitiveType, ToType> = {
    readonly to: ToType;
    readonly from: FromType;
};
export type CustomConversion<FromType extends PrimitiveType, ToType> = {
    readonly to: (val: FromType) => ToType;
    readonly from: (val: ToType) => FromType;
};
interface LayoutItemBase<BL extends BinaryLiterals> {
    readonly name: string;
    readonly binary: BL;
}
interface PrimitiveFixedCustom<T extends PrimitiveType> {
    custom: T;
    omit?: boolean;
}
interface OptionalToFromCustom<T extends PrimitiveType> {
    custom?: FixedConversion<T, any> | CustomConversion<T, any>;
}
interface UintLayoutItemBase<T extends number | bigint> extends LayoutItemBase<"uint"> {
    size: T extends bigint ? number : NumberSize;
}
export interface PrimitiveFixedUintLayoutItem<T extends number | bigint> extends UintLayoutItemBase<T>, PrimitiveFixedCustom<T> {
}
export interface OptionalToFromUintLayoutItem<T extends number | bigint> extends UintLayoutItemBase<T>, OptionalToFromCustom<T> {
}
export interface FixedPrimitiveBytesLayoutItem extends LayoutItemBase<"bytes">, PrimitiveFixedCustom<Uint8Array> {
}
export interface FixedValueBytesLayoutItem extends LayoutItemBase<"bytes"> {
    readonly custom: FixedConversion<Uint8Array, any>;
}
export interface FixedSizeBytesLayoutItem extends LayoutItemBase<"bytes"> {
    readonly size: number;
    readonly custom?: CustomConversion<Uint8Array, any>;
}
export interface LengthPrefixedBytesLayoutItem extends LayoutItemBase<"bytes"> {
    readonly lengthSize?: NumberSize;
    readonly custom?: CustomConversion<Uint8Array, any>;
}
export interface ArrayLayoutItem extends LayoutItemBase<"array"> {
    readonly lengthSize?: NumberSize;
    readonly layout: Layout;
}
export interface ObjectLayoutItem extends LayoutItemBase<"object"> {
    readonly layout: Layout;
}
export type UintLayoutItem = PrimitiveFixedUintLayoutItem<number> | OptionalToFromUintLayoutItem<number> | PrimitiveFixedUintLayoutItem<bigint> | OptionalToFromUintLayoutItem<bigint>;
export type BytesLayoutItem = FixedPrimitiveBytesLayoutItem | FixedValueBytesLayoutItem | FixedSizeBytesLayoutItem | LengthPrefixedBytesLayoutItem;
export type LayoutItem = UintLayoutItem | BytesLayoutItem | ArrayLayoutItem | ObjectLayoutItem;
export type Layout = readonly LayoutItem[];
type NameOrOmitted<T extends {
    name: PropertyKey;
}> = T extends {
    omit: true;
} ? never : T["name"];
export type LayoutToType<L extends Layout> = {
    readonly [I in L[number] as NameOrOmitted<I>]: LayoutItemToType<I>;
};
export type LayoutItemToType<I extends LayoutItem> = [
    I
] extends [ArrayLayoutItem] ? LayoutToType<I["layout"]>[] : [I] extends [ObjectLayoutItem] ? LayoutToType<I["layout"]> : [I] extends [UintLayoutItem] ? I["custom"] extends number | bigint ? I["custom"] : I["custom"] extends CustomConversion<any, infer ToType> ? ToType : I["custom"] extends FixedConversion<any, infer ToType> ? ToType : UintSizeToPrimitive<I["size"]> : [I] extends [BytesLayoutItem] ? I["custom"] extends CustomConversion<any, infer ToType> ? ToType : I["custom"] extends FixedConversion<any, infer ToType> ? ToType : Uint8Array : never;
export {};
//# sourceMappingURL=layout.d.ts.map