"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.constMap = void 0;
//everything in here is really just a band-aid/cover for the fact that TypeScript (despite having
// capital T type in its name) lacks support for higher-order type functions.
//find tests/examples at the bottom to get some idea wtf is going on here
const array_1 = require("./array");
function isMappableKey(key) {
    return ["string", "number", "symbol", "bigint", "boolean"].includes(typeof key);
}
const isRecursiveTuple = (arr) => arr.length === 2 && !Array.isArray(arr[0]) && Array.isArray(arr[1]);
const cartesianRightRecursive = (arr) => (arr.length === 0
    ? []
    : Array.isArray(arr[0])
        ? arr.map(([key, val]) => Array.isArray(val)
            ? (isRecursiveTuple(val) ? cartesianRightRecursive(val) : val).map(ele => [key, ele].flat())
            : [[key, val]]).flat()
        : isRecursiveTuple(arr)
            ? cartesianRightRecursive(arr[1]).map((ele) => [arr[0], ele])
            : arr);
const toMapping = (mapping, shape) => {
    const crr = cartesianRightRecursive(mapping);
    if (crr.length === 0)
        throw new Error("Invalid mapping: empty");
    const definedShape = (shape === undefined)
        ? [(0, array_1.range)(crr[0].length - 1), [crr[0].length - 1]]
        : shape.map(ind => typeof ind === "number" ? [ind] : ind);
    let leafObjects = [];
    //if our leafs are arrays to begin with then we should not strip them
    let allSingletons = definedShape[1].length === 1;
    const buildMapping = (keyCartesianSet, values) => {
        const distinctKeys = Array.from(new Set(keyCartesianSet[0]).values());
        const keyRows = new Map(distinctKeys.map(key => [key, []]));
        for (const [i, key] of keyCartesianSet[0].entries())
            keyRows.get(key).push(i);
        if (keyCartesianSet.length === 1) {
            const ret = Object.fromEntries(distinctKeys.map(key => [key, keyRows.get(key).flatMap(i => values[i])]));
            if (allSingletons) {
                for (const valRow of keyRows.values())
                    if (valRow.length !== 1) {
                        allSingletons = false;
                        return ret;
                    }
                leafObjects.push(ret);
            }
            return ret;
        }
        const droppedKeyCol = (0, array_1.zip)(keyCartesianSet.slice(1));
        return Object.fromEntries(distinctKeys.map(key => {
            const rows = keyRows.get(key);
            const keyCartesianSubset = (0, array_1.zip)(rows.map(i => droppedKeyCol[i]));
            const valuesSubset = rows.map(i => values[i]);
            return [key, buildMapping(keyCartesianSubset, valuesSubset)];
        }));
    };
    const cols = (0, array_1.zip)(crr);
    const getCol = (col) => {
        const colArr = cols[col];
        if (colArr === undefined)
            throw new Error(`Invalid shape: column ${col} does not exist`);
        return colArr;
    };
    const [keyCartesianSet, leafValues] = definedShape.map(indx => indx.map(col => getCol(col)));
    if (keyCartesianSet.length === 0)
        throw new Error("Invalid shape: empty key set");
    if (leafValues.length === 0)
        throw new Error("Invalid shape: empty value set");
    for (const keyCol of keyCartesianSet)
        for (const key of keyCol)
            if (!isMappableKey(key))
                throw new Error(`Invalid key: ${key} in ${keyCol}`);
    const ret = buildMapping(keyCartesianSet, (0, array_1.zip)(leafValues));
    if (allSingletons)
        for (const leafObj of leafObjects)
            for (const key of Object.keys(leafObj))
                leafObj[key] = leafObj[key][0];
    return ret;
};
const has = (f) => (...args) => f(...args) !== undefined;
const get = (f) => (...args) => f(...args);
const constMap = (mappingEntries, shape) => {
    const mapping = toMapping(mappingEntries, shape);
    const genericMappingFunc = ((...args) => args.reduce((subMapping, key) => subMapping ? subMapping[key.toString()] ?? undefined : undefined, mapping));
    genericMappingFunc["get"] = get(genericMappingFunc);
    genericMappingFunc["has"] = has(genericMappingFunc);
    return genericMappingFunc;
};
exports.constMap = constMap;
//--- find a bunch of "tests" below
//It's unclear to me how to even properly test Typescript types, seeing how there isn't even an
//  equals operation...
//So this will have to do for now...
// const sample = [
//   [
//     "Mainnet", [
//       ["Ethereum", 1n],
//       ["Bsc", 56n],
//     ]
//   ],
//   [
//     "Testnet", [
//       ["Ethereum", 5n],
//       ["Sepolia", 11155111n],
//     ]
//   ]
// ] as const satisfies MappingEntries;
// type Test1 = ToMapping<typeof sample>;
// type Test2 = ToMapping<typeof sample, [[0,1],2]>; //same as Test1
// type Test3 = ToMapping<typeof sample, [[0,1],[2]]>; //same as Test1
// type Test10 = ToMapping<typeof sample, [[0,1],[0,1,2]]>;
// type Test11 = ToMapping<typeof sample, [[0,1],[2,1,0]]>;
// type Test20 = ToMapping<typeof sample, [0,1]>;
// type Test21 = ToMapping<typeof sample, [[0],1]>; //same as Test20
// type Test22 = ToMapping<typeof sample, [0,[1]]>; //same as Test20
// type Test23 = ToMapping<typeof sample, [[0],[1]]>; //same as Test20
// type Test30 = ToMapping<typeof sample, [2,0]>;
// type Test31 = ToMapping<typeof sample, [2,[0,1]]>;
// type Test32 = ToMapping<typeof sample, [[1,0],2]>;
// type Test40 = ToMapping<typeof sample, [1,0]>;
// const test1 = constMap(sample);
// const test1Entry1 = test1("Testnet", "Sepolia"); //11155111n
// const test2 = constMap(sample, [[0,1],2]); //same as test1
// const test2Entry1 = test2("Testnet", "Sepolia"); //same as test1Entry1
// const test2Entry2 = test2.get("doesn't", "exist"); //undefined: bigint | undefined
// const test2Entry3 = test2.has("doesn't", "exist"); //false: boolean
// const test10 = constMap(sample, [[0,1],[0,1,2]]);
// const test10Entry1 = test10("Testnet", "Sepolia"); //["Testnet", "Sepolia", 11155111n]
// const test20 = constMap(sample, [0,1]);
// const test20Entry1 = test20("Testnet"); //["Ethereum", "Sepolia"]
// const test30 = constMap(sample, [2,0]);
// const test30Entry1 = test30(1n); //"Mainnet"
// const test31 = constMap(sample, [2,[0,1]]);
// const test31Entry1 = test31(1n); //["Mainnet", "Ethereum"]
// const test31Entry2 = test31(11155111n); //["Testnet", "Sepolia"]
// const test40 = constMap(sample, [1,0]);
// const test40Entry1 = test40("Ethereum"); //["Mainnet", "Testnet"]
// const test40Entry2 = test40("Sepolia"); //["Testnet"]
// const test40Entry3 = test40("Bsc"); //["Mainnet"]
//# sourceMappingURL=mapping.js.map