import { isPrimitiveType, } from "./layout";
export const fixedItemsOfLayout = (layout) => layout.reduce((acc, item) => {
    if (item["custom"] !== undefined && (isPrimitiveType(item["custom"]) || isPrimitiveType(item["custom"].from)))
        return [...acc, item];
    if (item.binary === "array" || item.binary === "object") {
        const fixedItems = fixedItemsOfLayout(item.layout);
        if (fixedItems.length > 0)
            return [...acc, { ...item, layout: fixedItems }];
    }
    return acc;
}, []);
export const dynamicItemsOfLayout = (layout) => layout.reduce((acc, item) => {
    if (item["custom"] === undefined || !(isPrimitiveType(item["custom"]) || isPrimitiveType(item["custom"].from)))
        return [...acc, item];
    if (item.binary === "array" || item.binary === "object") {
        const dynamicItems = dynamicItemsOfLayout(item.layout);
        if (dynamicItems.length > 0)
            return [...acc, { ...item, layout: dynamicItems }];
    }
    return acc;
}, []);
export const addFixedValues = (layout, dynamicValues) => {
    const ret = {};
    for (const item of layout) {
        if (item.binary === "object") {
            const subDynamicValues = ((item.name in dynamicValues)
                ? dynamicValues[item.name]
                : {});
            ret[item.name] = addFixedValues(item.layout, subDynamicValues);
        }
        else if (item.binary === "array") {
            const subDynamicValues = ((item.name in dynamicValues)
                ? dynamicValues[item.name]
                : []);
            ret[item.name] = subDynamicValues.map(element => addFixedValues(item.layout, element));
        }
        else if (item.custom !== undefined &&
            (isPrimitiveType(item.custom) || isPrimitiveType(item.custom.from))) {
            if (!item?.omit)
                ret[item.name] = isPrimitiveType(item.custom) ? item.custom : item.custom.to;
        }
        else
            ret[item.name] = dynamicValues[item.name];
    }
    return ret;
};
//# sourceMappingURL=fixedDynamic.js.map