import { IndexEs, Flatten, Unflatten, InnerFlatten, IsRectangular, Zip, Cartesian, OnlyIndexes, ExcludeIndexes, Entries, Column } from "./array";
import { AssertType, Function, Widen, RoArray, RoArray2D } from "./metaprogramming";
export type ShallowMapping<M extends RoArray<readonly [PropertyKey, unknown]>> = {
    readonly [E in M[number] as E[0]]: E[1];
};
type MappableKey = PropertyKey | bigint | boolean;
type ToExtPropKey<T> = T extends MappableKey ? AssertType<T extends bigint ? `bigint(${T})` : T extends boolean ? `boolean(${T})` : T, PropertyKey> : never;
type FromExtPropKey<T extends PropertyKey> = T extends `bigint(${infer V extends bigint})` ? V : T extends `boolean(${infer V extends boolean})` ? V : T;
type MappingEntry = readonly [MappableKey, unknown];
type MappingEntries = RoArray<MappingEntry>;
type CombineKeyValues<K, T extends MappingEntries, M extends RoArray, U extends MappingEntries = []> = T extends readonly [infer Head extends MappingEntry, ...infer Tail extends MappingEntries] ? Head extends readonly [infer Key extends MappableKey, infer Value] ? Key extends K ? CombineKeyValues<K, Tail, [...M, Value], U> : CombineKeyValues<K, Tail, M, [...U, Head]> : never : [M, U];
type ToMapEntries<T extends RoArray2D, M extends MappingEntries = []> = T extends readonly [infer Head, ...infer Tail extends MappingEntries] ? Head extends readonly [infer Key extends MappableKey, infer Value] ? CombineKeyValues<Key, Tail, [Value]> extends readonly [
    infer MK extends RoArray,
    infer MU extends RoArray2D
] ? ToMapEntries<MU, [...M, [Key, MK]]> : never : never : M;
type CartesianRightRecursive<T extends RoArray> = T extends RoArray<readonly [MappableKey, RoArray]> ? Flatten<[
    ...{
        [K in keyof T]: K extends `${number}` ? InnerFlatten<Cartesian<T[K][0], CartesianRightRecursive<T[K][1]>>> : never;
    }
]> : T extends readonly [MappableKey, RoArray] ? Cartesian<T[0], T[1]> : T;
type Shape = readonly [IndexEs, IndexEs];
type CartesianSet<T = unknown> = RoArray2D<T>;
type Transpose<T extends RoArray2D> = Zip<T>;
type DivideAndConquer<KA extends CartesianSet<MappableKey>, VA extends RoArray, KeyRows extends RoArray<readonly [MappableKey, RoArray<number>]>> = [
    ...{
        [K in keyof KeyRows]: [
            KeyRows[K][0],
            ProcessNextKey<OnlyIndexes<KA, KeyRows[K][1]>, OnlyIndexes<VA, KeyRows[K][1]>>
        ];
    }
];
type ProcessNextKey<KA extends CartesianSet<MappableKey>, VA extends RoArray> = KA["length"] extends 0 ? VA : DivideAndConquer<Transpose<ExcludeIndexes<Transpose<KA>, 0>>, VA, ToMapEntries<Entries<Column<KA, 0>>>>;
type SplitAndReorderKeyAndValueColums<A extends CartesianSet, S extends Shape> = ProcessNextKey<Transpose<OnlyIndexes<Transpose<A>, S[0]>>, Unflatten<S[1] extends number | readonly [number] ? Transpose<OnlyIndexes<Transpose<A>, S[1]>> : Unflatten<Transpose<OnlyIndexes<Transpose<A>, S[1]>>>>>;
export type LeafValue = readonly [readonly [unknown]];
type UnwrapValuesIfAllAreSingletons<M extends MappingEntries> = M extends readonly [infer Head, ...infer Tail extends MappingEntries] ? Head extends readonly [infer K, infer V] ? V extends readonly [LeafValue, LeafValue] ? false : V extends MappingEntries ? UnwrapValuesIfAllAreSingletons<V> extends MappingEntries ? [
    [
        K,
        UnwrapValuesIfAllAreSingletons<V>
    ],
    ...UnwrapValuesIfAllAreSingletons<Tail>
] : false : V extends readonly [unknown] ? [readonly [K, V[0]], ...UnwrapValuesIfAllAreSingletons<Tail>] : false : never : [];
type MaybeUnwrapValuesIfAllAreSingletons<M> = AssertType<UnwrapValuesIfAllAreSingletons<AssertType<M, MappingEntries>> extends false ? M : UnwrapValuesIfAllAreSingletons<AssertType<M, MappingEntries>>, MappingEntries>;
type TransformMapping<M extends MappingEntries, S extends Shape | undefined = undefined> = IsRectangular<CartesianRightRecursive<M>> extends true ? [S] extends [undefined] ? M : MaybeUnwrapValuesIfAllAreSingletons<SplitAndReorderKeyAndValueColums<CartesianRightRecursive<M>, Exclude<S, undefined>>> : never;
export type ObjectFromMappingEntries<M extends MappingEntries> = {
    [K in keyof M as (K extends `${number}` ? ToExtPropKey<M[K][0]> : never)]: M[K][1] extends readonly [LeafValue, LeafValue] ? Flatten<Flatten<M[K][1]>> : M[K][1] extends MappingEntries ? ObjectFromMappingEntries<M[K][1]> : M[K][1] extends RoArray<LeafValue> ? Flatten<Flatten<M[K][1]>> : M[K][1] extends LeafValue ? M[K][1][0][0] : M[K][1];
};
export type ToMapping<M extends MappingEntries, S extends Shape | undefined = undefined> = ObjectFromMappingEntries<TransformMapping<M, S>>;
type Mapped = {
    [key: PropertyKey]: unknown | Mapped;
};
type RecursiveAccess<M extends Mapped, KA extends RoArray<MappableKey>> = KA extends readonly [infer Head extends MappableKey, ...infer Tail extends RoArray<MappableKey>] ? ToExtPropKey<Head> extends keyof M ? M[ToExtPropKey<Head>] extends Mapped ? RecursiveAccess<M[ToExtPropKey<Head>], Tail> : M[ToExtPropKey<Head>] : never : M;
type GenericMappingFunc<M extends Mapped, D extends number> = D extends 1 ? <K1 extends FromExtPropKey<keyof M>>(...args: [K1]) => RecursiveAccess<M, [K1]> : D extends 2 ? <K1 extends FromExtPropKey<keyof M>, K2 extends FromExtPropKey<keyof RecursiveAccess<M, [K1]>>>(...args: [K1, K2]) => RecursiveAccess<M, [K1, K2]> : D extends 3 ? <K1 extends FromExtPropKey<keyof M>, K2 extends FromExtPropKey<keyof RecursiveAccess<M, [K1]>>, K3 extends FromExtPropKey<keyof RecursiveAccess<M, [K1, K2]>>>(...args: [K1, K2, K3]) => RecursiveAccess<M, [K1, K2, K3]> : D extends 4 ? <K1 extends FromExtPropKey<keyof M>, K2 extends FromExtPropKey<keyof RecursiveAccess<M, [K1]>>, K3 extends FromExtPropKey<keyof RecursiveAccess<M, [K1, K2]>>, K4 extends FromExtPropKey<keyof RecursiveAccess<M, [K1, K2, K3]>>>(...args: [K1, K2, K3, K4]) => RecursiveAccess<M, [K1, K2, K3, K4]> : never;
type MappingDepth<M extends MappingEntries, S extends Shape | undefined = undefined> = CartesianRightRecursive<TransformMapping<M, S>> extends RoArray2D ? ExcludeIndexes<CartesianRightRecursive<TransformMapping<M, S>>[0], 0>["length"] : 0;
type WidenArray<T extends RoArray> = T extends readonly [infer Head, ...infer Tail] ? [Widen<Head>, ...WidenArray<Tail>] : [];
type WidenParams<F extends Function> = WidenArray<Parameters<F>>;
type Has<F extends Function> = Function<WidenParams<F>, boolean>;
type Get<F extends Function> = Function<WidenParams<F>, ReturnType<F> | undefined>;
type ToGenericMappingFunc<M extends MappingEntries, S extends Shape | undefined = undefined> = GenericMappingFunc<ToMapping<M, S>, MappingDepth<M, S>>;
type ConstMapRet<F extends Function> = F & {
    get: Get<F>;
    has: Has<F>;
};
export declare const constMap: <const M extends MappingEntries, const S extends Shape | undefined = undefined>(mappingEntries: M, shape?: S | undefined) => ConstMapRet<GenericMappingFunc<ObjectFromMappingEntries<TransformMapping<M, S>>, MappingDepth<M, S>>>;
export {};
//# sourceMappingURL=mapping.d.ts.map